#include <asm/segment.h>
#include "videodev.h"
#include "vidcap.h"
#include "igs.h"


/* lookup tables for color conversion */
int XlatY[XLATTABSIZE] =
{0};
int XlatU_B[XLATTABSIZE] =
{0};
int XlatU_G[XLATTABSIZE] =
{0};
int XlatV_G[XLATTABSIZE] =
{0};
int XlatV_R[XLATTABSIZE] =
{0};


/**
 * Color translation table used for YUV to RGB conversion.
 */
#define ORIG_XLAT 1
void
InitXlatTab (void)			/*YUV->RGB xlat table */
{
  int i, j;

  for (i = 0; i < XLATTABSIZE; i++)
  {
#ifdef ORIG_XLAT
	  if (i < 16)
		j = 16;
	  else if (i > 253)
		j = 253;
          else
		j = i;
#else
          j = (255 * i + 110) / 220; // scale up
          if (j>255) j=255;
#endif

	  // orig: XlatY[i] = (int ) j;
	  XlatY[i] = j;
	}

  for (i = 0; i < XLATTABSIZE; i++)
  {
#ifdef ORIG_XLAT
	  if (i < 16)
		j = 16;
	  else if (i > 240)
		j = 240;
	  else
		j = i;

          j -= 128;
#else
          j = i - 128;  // make signed
          if (j<0) j++; // noise reduction
          j = (127 * j + 56) / 112; // scale up
          if (j<-128) j=-128;
          if (j>127) j=127;
          
#endif

	  XlatU_B[i] = MulDiv (j, 219, 126);	/* j*219/126 */
	  XlatU_G[i] = MulDiv (j, 74, 219);
	  XlatV_G[i] = MulDiv (j, 153, 219);
	  XlatV_R[i] = MulDiv (j, 219, 160);
	}
}


/***************************************************************
Convert a single line of YUV 4:2:2 data to 24 bpp RGB data.

given YCrCb in the range 16...235 (Y=Y,Cr=V,Cb=U)
|R| = | 1  1.371  0	| |Y    |
|G| = | 1 -0.698 -0.336 | |V-128|
|B| = | 1  0	  1.732 | |U-128|

given YCrCb in the range 0..255 (Y=Y,Cr=V,Cb=U)
|R| = | 1.164  1.596  0	    | |Y -16|
|G| = | 1.164 -0.813 -0.392 | |V-128|
|B| = | 1.164  0      2.017 | |U-128|

in:
	lpRGBBuf	- Pointer to the RGB destination buffer
	lpYUVBuf	- Pointer to the YUV source buffer
	nLineLen	- Length of the YUV line in pixels (must <=1048)
***************************************************************/
void
YUV422To24RGB (unsigned char *lpRGBBuf, unsigned char *lpYUVBuf, int nLineLen)
{
  int i;
  int *lpnYUVBuf;
  unsigned char *lpbRGBBuf;
  unsigned char y0, uu, y1, vv;
  int yuv;
  unsigned char data = 0;

  lpnYUVBuf = (unsigned int*) lpYUVBuf;
  lpbRGBBuf = lpRGBBuf;

  for (i = 0; i < nLineLen; i += 4)
  {
      /* !!! WHATS GOING ON HERE !!!!
       If I take the byte order as Y0, U, Y1, V
       then I get bad colors; If I take the byte order as Y0, V, Y1, U,
       then I get good colors ???? */
      yuv = *lpnYUVBuf++;
      y0 = (unsigned char) yuv;
      vv = (unsigned char) (yuv >> 8);
      y1 = (unsigned char) (yuv >> 16);
      uu = (unsigned char) (yuv >> 24);

#if 0
	  /* 1st RGB pixel */
	  *lpbRGBBuf++ = (unsigned char)
		min (CVHI, max (CVLO, (XlatY[y0] + XlatU_B[uu])));
	  *lpbRGBBuf++ = (unsigned char)
		min (CVHI, max (CVLO, (XlatY[y0] - XlatU_G[uu] - XlatV_G[vv])));
	  *lpbRGBBuf++ = (unsigned char)
		min (CVHI, max (CVLO, (XlatY[y0] + XlatV_R[vv])));
#endif

	  /* 1st RGB pixel */
	  data = min (CVHI, max (CVLO, (XlatY[y0] + XlatU_B[uu])));
	  put_user_byte (data, lpbRGBBuf);
	  lpbRGBBuf++;				/* Blue */

	  data = min (CVHI, max (CVLO, (XlatY[y0] - XlatU_G[uu] - XlatV_G[vv])));
	  put_user_byte (data, lpbRGBBuf);
	  lpbRGBBuf++;				/* Green */

	  data = min (CVHI, max (CVLO, (XlatY[y0] + XlatV_R[vv])));
	  put_user_byte (data, lpbRGBBuf);
	  lpbRGBBuf++;				/* Red */

#if 0
	  /* 2nd RGB pixel */
	  *lpbRGBBuf++ = (unsigned char)
		min (CVHI, max (CVLO, (XlatY[y1] + XlatU_B[uu])));
	  *lpbRGBBuf++ = (unsigned char)
		min (CVHI, max (CVLO, (XlatY[y1] - XlatU_G[uu] - XlatV_G[vv])));
	  *lpbRGBBuf++ = (unsigned char)
		min (CVHI, max (CVLO, (XlatY[y1] + XlatV_R[vv])));
#endif
	  /* 2nd RGB pixel */
	  data = min (CVHI, max (CVLO, (XlatY[y1] + XlatU_B[uu])));
	  put_user_byte (data, lpbRGBBuf);
	  lpbRGBBuf++;				/* Blue */

	  data = min (CVHI, max (CVLO, (XlatY[y1] - XlatU_G[uu] - XlatV_G[vv])));
	  put_user_byte (data, lpbRGBBuf);
	  lpbRGBBuf++;				/* Green */

	  data = min (CVHI, max (CVLO, (XlatY[y1] + XlatV_R[vv])));
	  put_user_byte (data, lpbRGBBuf);
	  lpbRGBBuf++;				/* Red */
	}
}

/**********************************************************
Color conversion from YUV to RGB24
It also dynamically flips the image

in:
	fpDst   - far Ptr to Destination buffer (RGB buffer)
	fpSrc   - far Ptr to Source buffer (YUV buffer)
	wWidth  - Width of the rectangle (pixel based)
	wHeight - Height of the rectangle
**********************************************************/
void
ConvertYUVtoRGB24 (char *fpDst, char *fpSrc, unsigned short wWidth, unsigned short wHeight)
{
    int row;
    unsigned char *fpTmp;

    for (row = 0; row < (int) wHeight; row++)
    {
        fpTmp = (unsigned char *) fpSrc + ((long) wWidth * 2 * row);
        YUV422To24RGB (
                       /**
                        * uncomment to flip the image
                        * (unsigned char*)fpDst + ((long)wWidth*3 * (wHeight - (row + 1))),
                        */
                       (unsigned char *) fpDst + ((long) wWidth * 3 * row),	/* 3 bytes/pixel! */
                       fpTmp, wWidth * 2);
    }
}
/**********************************************************
Color conversion from YUV to RGB24
It also dynamically flips the image

in:
	fpDst   - far Ptr to Destination buffer (RGB buffer)
	fpSrc   - far Ptr to Source buffer (YUV buffer)
	wWidth  - Width of the rectangle (pixel based)
	wHeight - Height of the rectangle
**********************************************************/
void
ConvertYUVtoRGB16 (char *fpDst, char *fpSrc, unsigned short wWidth, unsigned short wHeight)
{
    int row;
    unsigned char *fpTmp;

    for (row = 0; row < (int) wHeight; row++)
    {
        fpTmp = (unsigned char *) fpSrc + ((long) wWidth * 2 * row);
        YUV422To16RGB (
                       /**
                        * uncomment to flip the image
                        * (unsigned char*)fpDst + ((long)wWidth*3 * (wHeight - (row + 1))),
                        */
                       (unsigned char *) fpDst + ((long) wWidth * 2 * row),	/* 2 bytes/pixel! */
                       fpTmp, wWidth * 2);
    }
}

void
YUV422To16RGB (unsigned char *lpRGBBuf,
               unsigned char *lpYUVBuf,
               int nLineLen)
{
  int i;
  unsigned long *lpnYUVBuf;
  unsigned char y0, uu, y1, vv;
  int yuv;
  unsigned long d_out;
  unsigned char data = 0;
  unsigned short d;

  lpnYUVBuf = (unsigned long *) lpYUVBuf;

  for (i = 0; i < nLineLen; i += 4)
  {
      yuv = *lpnYUVBuf++;			/*Y0 U0 */
      y0 = (unsigned char) yuv;
      vv = (unsigned char) (yuv >> 8);
      y1 = (unsigned char) (yuv >> 16);
      uu = (unsigned char) (yuv >> 24);


      /* 1st RGB pixel */
      data = min (CVHI, max (CVLO, (XlatY[y0] + XlatU_B[uu])));
      d = (data & 0xF8) << 8;

      data = min (CVHI, max (CVLO, (XlatY[y0] - XlatU_G[uu] - XlatV_G[vv])));
      d |= (data & 0xFC) << 3;

      data = min (CVHI, max (CVLO, (XlatY[y0] + XlatV_R[vv])));
      d_out = d | (data & 0xF8) >> 3;

      /* 2nd RGB pixel */
      data = min (CVHI, max (CVLO, (XlatY[y1] + XlatU_B[uu])));
      d = (data & 0xF8) << 8;

      data = min (CVHI, max (CVLO, (XlatY[y1] - XlatU_G[uu] - XlatV_G[vv])));
      d |= (data & 0xFC) << 3;

      data = min (CVHI, max (CVLO, (XlatY[y1] + XlatV_R[vv])));
      d_out |= (d | (data & 0xF8) >> 3) << 16;
      memcpy_tofs(lpRGBBuf, &d_out, 4);
      lpRGBBuf += 4;
  }
}
